// startrek_console
//
// a console-only version of startrek, for experiment
//
// usage: startrek [port#]
//
// to begin, this programme speaks only to the fiat 16V NA, and it
// assumes the use of a 409.1 KKL type adaptor (i.e. commands are 
// sent to the K line and echoed on the L line; returned values
// are on the L line.)
// it does not try and enumerate a list of ports; it tries to open 
// the serial port specified in the command line and if it cannot
// open then the programme quits
//
// Released under LGPL licence; see www.gnu.org for details
// (c) neil barnes/nailed_barnacle 2009
// nailed_barnacle@hotmail.com
//
// compiles under Visual Studio VC++ 2005

// first we do away with MS non-standard warning crap
// if I write unsafe code you may assume that I want to

#define _CRT_SECURE_NO_DEPRECATE

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <windows.h>

// global variables... how bad is that?
// these are for the serial code
static HANDLE					rs232;
static DCB						dcb;
static DWORD					txcount,rxcount;
static COMMTIMEOUTS		cto;

// some colour definitions
#define RED (FOREGROUND_RED | FOREGROUND_INTENSITY)
#define GREEN (FOREGROUND_GREEN | FOREGROUND_INTENSITY)
#define BLUE (FOREGROUND_BLUE | FOREGROUND_INTENSITY)
#define WHITE (FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY)
#define YELLOW (FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_INTENSITY)


HANDLE serial_setcomms (int baud, char * comport);
unsigned char	GetECUByte (unsigned char ch);
void PrintColour(HANDLE con, char * str, int x, int y, WORD colour);

// temperature conversion look up table
int temperature[256] = {
	125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,
	125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,
	125,122,119,116,113,111,108,105,102, 99, 96, 93, 90, 87, 84, 82,
	 79, 77, 76, 75, 74, 73, 71, 70, 69, 68, 66, 65, 64, 63, 62, 60,
	 59, 58, 58, 57, 56, 56, 55, 54, 53, 53, 52, 51, 50, 50, 49, 48,
	 48, 47, 46, 46, 45, 45, 44, 43, 43, 42, 42, 41, 40, 40, 39, 39,
	 38, 37, 37, 36, 36, 35, 35, 34, 34, 33, 33, 32, 32, 31, 31, 30,
	 30, 29, 29, 28, 28, 27, 27, 26, 26, 25, 25, 24, 24, 23, 23, 22,
	 22, 22, 21, 21, 20, 20, 19, 19, 18, 18, 17, 17, 16, 16, 15, 15,
	 14, 14, 13, 13, 12, 12, 11, 11, 10, 10,  9,  9,  8,  8,  7,  7,
	  6,	6,  5,  5,  4,  4,  3,  3,  2,  2,  1,  1,  0,  0, -1, -1,
	 -2, -2, -3, -4, -4, -5, -6, -6, -7, -7, -8, -9, -9,-10,-11,-11,
	-12,-13,-14,-15,-16,-17,-18,-19,-20,-21,-22,-23,-24,-26,-27,-28,
	-29,-34,-35,-37,-39,-40,-42,-43,-45,-47,-48,-50,-52,-53,-55,-57,
	-55,-55,-55,-55,-55,-55,-55,-55,-55,-55,-55,-55,-55,-55,-55,-55,
	-55,-55,-55,-55,-55,-55,-55,-55,-55,-55,-55,-55,-55,-55,-55,-55 };


int main (int argc, char * argv[])
{
char tmp[128];		// temporary store for short messages
char kb;					// last keyboard hit
int ecu;					// the byte or word returned by the ecu, 
									// before processing into real values
char uniram1;
char uniram2;
char uniram3;			// the error bytes
	
float fecu;				// floating point converted value
int tumbler;			// state for the \|/- tumbler
int turbo = 0;		// 0 for NA engine, 1 for turbo engine

HANDLE con;				// the pointer to the active console
COORD	posn;				// coords for cursor
									
	// start by getting a handle to the console
	con = CreateFile(	"CONOUT$",
										GENERIC_READ | GENERIC_WRITE,
										0,
										NULL,
										OPEN_EXISTING,
										FILE_SHARE_WRITE,
										NULL);

	// label it
	SetConsoleTitle("Star Trek console version 1.0 - IAW P8 PQP");
	
	// clear the screen (we might have been opened from cmd
	for (int q=0; q<25; q++)
		PrintColour(con,"                                                                               ",0,q,WHITE);
	PrintColour(con,"",0,0,WHITE);
	
	if (argc < 2)
	{
		printf("Usage: startrek [port#]\n");
		printf("You didn't specify the port, so assuming com1\n");
	}
	SetConsoleTextAttribute(con,WHITE);
	printf("Star Trek console version 1.0\n");
	printf("(c) neil barnes/nailed_barnacle 2009\n");
	
	// first we try and open the requested port, or we do nothing...
	// we must convert the port - presented as a number - to the expected
	// text format \\.\comxx
	if (argc>1)
		sprintf(tmp,"\\\\.\\com%s",argv[1]);
	else
		sprintf(tmp,"\\\\.\\.com1");
	printf("Attempting to open the comms port '%s'...\n",tmp);
	if (serial_setcomms (7812,tmp) == 0)
	{
		printf("Unable to open serial port; quiting.\n");
		exit(1);
	}
	else
		printf("Successfully opened port!\n\n");
	// wait a little while and then erase the message
	SleepEx(2000, false);	
	posn.X = 0;
	posn.Y = 2;
	SetConsoleCursorPosition(con,posn);
	printf("																						    \n");
	printf("                                                \n");
	PrintColour(con,"Hit 'q' to quit",0,3,YELLOW);
	PrintColour(con,"Turbo mode? (hit 't') [NO ]",40,3,YELLOW);
	
	// if we had any intialisation to do, this is where we'd do it...
	
	// we can write the ECU ID data once only
	// <iso location="48" length="6" mode="hex"/>
	// <spare location="54" length="11" mode="ascii"/>
	
	for (int q=48; q<54; q++)
		tmp[q-48] = GetECUByte(q);
	PrintColour(con,"ECU ID: ",0,21,GREEN);
	printf("%0.2X%0.2X%0.2X%0.2X%0.2X%0.2X",(unsigned char)tmp[0],(unsigned char)tmp[1],(unsigned char)tmp[2],(unsigned char)tmp[3],(unsigned char)tmp[4],(unsigned char)tmp[5]);

	sprintf(tmp,"Spare code: ");
	for (int q=54; q<66; q++)
		tmp[q-54+12] = GetECUByte(q);
	tmp[24] = '\0';
	PrintColour(con,tmp,40,21,GREEN);
	
	// and the error labels are only written once
	PrintColour(con,"Idle control valve", 0,11,BLUE);
	PrintColour(con,"Mixture ratio limit", 40,11,BLUE);
	PrintColour(con,"Air temp sensor", 0,12,BLUE);
	PrintColour(con,"Knock (pinking)", 40,12,BLUE);
	PrintColour(con,"Water temp sensor", 0,13,BLUE);
	PrintColour(con,"Lambda sensor", 40,13,BLUE);
	PrintColour(con,"MAP sensor", 0,14,BLUE);
	PrintColour(con,"Throttle posn sensor", 40,14,BLUE);
	PrintColour(con,"Camshaft sensor", 0,15,BLUE);
	PrintColour(con,"ECU EPROM", 40,15,BLUE);
	PrintColour(con,"ECU EEPROM", 0,16,BLUE);
	PrintColour(con,"ECU RAM", 40,16,BLUE);
	PrintColour(con,"Crankshaft sensor", 0,17,BLUE);
	PrintColour(con,"Immobiliser",0,19,BLUE);

	// now we sit in a loop querying the ECU, receiving the result, 
	// converting it to something human readable, and displaying it.
	
	// loop forever
	tumbler = 0;
	while (1)
	{
		if (_kbhit())
		{
			kb = _getch();
			switch (kb)
			{
				// quit on q key, either case
				case 'q':
				case 'Q':
					break;
				
				// we can drop in here any other keypresses
				// we might find useful later here
				
				// toggle turbo mode
				case 't':
				case 'T':
					turbo = 1-turbo;
					if (turbo)
						PrintColour(con,"YES",63,3,YELLOW);
					else
						PrintColour(con,"NO ",63,3,YELLOW);
					break;
					
				// and then the main loop	
				default:
					break;
			}
		}
		else
		{
			if ((kb == 'q') | (kb == 'Q'))
				break;
			
			// we implement a rotating whirly thing so we can see it's working
			switch (tumbler)
			{
				case 0:
					PrintColour(con,"\\",16,3,YELLOW);
					break;
				case 1:
					PrintColour(con,"|",16,3,YELLOW);
					break;
				case 2:
					PrintColour(con,"/",16,3,YELLOW);
					break;
				case 3:
					PrintColour(con,"-",16,3,YELLOW);
					break;
				default:
					break;
			}
			tumbler ++;
			tumbler &= 3;
			
			// in turn, we ask for the parameter from the ecu
			// apply the conversion factor, and display the result
			// most parameters are single bytes but some require two 
			// bytes; one or two require table lookup
			
			// engine period msb,lsb
			ecu = (GetECUByte(1)<<8)+GetECUByte(2);
			// 30,000,000/engine period = revs per minute
			if (ecu != 0)
				fecu = (float)30000000/ecu;
			else
				fecu = (float)0;
			sprintf(tmp,"Revs:\t\t%d rpm     ",(int)fecu);
			if ((fecu < 500) || (fecu > 6500))
				PrintColour(con,tmp,0,5,RED);
			else
				if ((fecu < 800) || (fecu > 6000))
					PrintColour(con,tmp,0,5,YELLOW);
				else
					PrintColour(con,tmp,0,5,GREEN);
						
			// injector period msb,lsb
			ecu = (GetECUByte(3)<<8)+GetECUByte(4);
			// 4* injector period = period in us
			fecu = (float)(ecu*4)/1000;  // we like ms
			sprintf(tmp,"Injectors:\t%2.1f ms     ",fecu);
			if (fecu < 0.01)
				PrintColour(con,tmp,40,5,RED);
			else
				PrintColour(con,tmp,40,5,GREEN);
			
			// advance
			ecu = GetECUByte(5);
			// advance/4 = degrees
			fecu = (float)ecu/4;
			sprintf(tmp,"Advance:\t%d degrees     ",(int)fecu);
			PrintColour(con,tmp,0,6,GREEN);
									
			// manifold pressure
			ecu = GetECUByte(6);
			if (turbo == 0)
				// manifold pressure for NA is vacuum*2.8758 + 91 mmHg
				// but millibar is more useful, and as one mb = 0.75mmHg
				// we can multiply the result by 4/3
				fecu = (float)((((float)ecu*2.8758)+91)*4/3);
			else
				// it's a turbo, the rules are different
				// now it's data*6.4161+45.63, mult by 4/3 to get millibar
				fecu = (float)((((float)ecu*6.4161)+45.63)*4/3);
			sprintf(tmp,"Manifold:\t%d mm Hg     ",(int)fecu);
			PrintColour(con,tmp,40,6,GREEN);
			
			// air and water temp
			ecu = GetECUByte(7);
			fecu = (float)temperature[ecu];
			sprintf(tmp,"Air temp:\t%d degrees     ",(int)fecu);
			PrintColour(con,tmp,0,7,GREEN);
			
			ecu = GetECUByte(8);
			fecu = (float)temperature[ecu];
			sprintf(tmp,"Water temp:\t%d degrees     ",(int)fecu);
			if (fecu > 110.0)
				PrintColour(con,tmp,40,7,RED);
			else
				if (fecu > 105.0)
					PrintColour(con,tmp,40,7,YELLOW);
				else
					PrintColour(con,tmp,40,7,GREEN);

			// throttle valve angle
			ecu = GetECUByte(9);
			// throttle valve angle is different depending on value
			// if angle is less than 30 degrees, angle is (value*0.1848)-1.41 degrees
			// if greater than 30 degrees, angle is (value*0.7058)-90 degrees
			// the turnover point is when the value is 171
			// this gives a much greater precision on small throttle openings
			if (ecu <171)
				fecu = (float)(((float)ecu*0.1848)-1.41);
			else
				fecu = (float)(((float)ecu*0.7058)-90);
			sprintf(tmp,"Throttle:\t%d degrees     ",(int)fecu);
			PrintColour(con,tmp,0,8,GREEN);
		
			// battery voltage
			ecu = GetECUByte(10);
			// value * 0.0628 volts
			fecu = (float)((float)ecu*0.0628);
			sprintf(tmp,"Battery:\t%2.2f volts     ",fecu);
			if ((fecu < 12.0) || (fecu > 14.5))
				PrintColour(con,tmp,40,8,RED);
			else
				if ((fecu < 12.5) || (fecu > 14.0))
					PrintColour(con,tmp,40,8,YELLOW);
				else
					PrintColour(con,tmp,40,8,GREEN);

			
			// injection timing angle
			ecu = GetECUByte(12);
			// multiply byte by 90/4, subtract from 720
			fecu = (float)(720-((float)ecu*90/4));
			sprintf(tmp,"Inj. angle:\t%d degrees     ",(int)fecu);
			PrintColour(con,tmp,0,9,GREEN);
			
			// VAE duty cycle (idle control valve)
			ecu = GetECUByte(13);
			// just multiply by 100 and divide by 255 for percentage
			fecu = (float)((float)ecu*100/255);
			sprintf(tmp,"ICR:\t\t%d percent     ",(int)fecu);
			PrintColour(con,tmp,40,9,GREEN);

			// now we get the temporary errors from ram
			// we print the error labels the same colour
			// but vary the colour of the message, if present
			// uniram 1, 2, 3

			uniram1 = GetECUByte(15);
			uniram2 = GetECUByte(16);
			uniram3 = GetECUByte(17);

			// icv
			if (uniram1 & 1)
				PrintColour(con,"Blocked",21,11,RED);
			else
				PrintColour(con,"OK     ",21,11,GREEN);
			
			// enrichment limit
			if (uniram1 & 2)
				if (uniram3 & 1)
					PrintColour(con,"Maximum lean",61,11,RED);
				else
						PrintColour(con,"Maximum rich",61,11,RED);			
			else
				PrintColour(con,"OK          ",61,11,GREEN);
			
			// air temp sensor
			if (uniram1 & 4)
				if (uniram3 & 2)
					PrintColour(con,"Short circuit",21,12,RED);
				else
					PrintColour(con,"Open circuit ",21,12,RED);
			else
				PrintColour(con,"OK           ",21,12,GREEN);
				
			// knock
			if (uniram1 & 8)
				PrintColour(con,"Pinking",61,12,RED);
			else
				PrintColour(con,"OK     ",61,12,GREEN);
				
			// water temp sensor
			if (uniram1 & 16)
				if (uniram3 & 4)
					PrintColour(con,"Short circuit",21,13,RED);
				else
					PrintColour(con,"Open circuit ",21,13,RED);
			else
				PrintColour(con,"OK           ",21,13,GREEN);
			
			// lambda sensor
			if (uniram1 & 32)
				PrintColour(con,"Faulty",61,13,RED);
			else
				PrintColour(con,"OK    ",61,13,GREEN);
				
			// map sensor
			if (uniram1 & 64)
				if (uniram3 & 128)
					PrintColour(con,"Short circuit",21,14,RED);
				else
					PrintColour(con,"Open circuit ",21,14,RED);
			else
				PrintColour(con,"OK           ",21,14,GREEN);
			
			// throttle position sensor
			if (uniram1 & 128)
				if (uniram3 & 16)
					PrintColour(con,"Short circuit",61,14,RED);
				else
					PrintColour(con,"Open circuit ",61,14,RED);
			else
				PrintColour(con,"OK           ",61,14,GREEN);
				
			// camshaft sensor
			if (uniram2 & 1)
				if (uniram3 & 128)
					PrintColour(con,"Short circuit",21,15,RED);
				else
					PrintColour(con,"Open circuit ",21,15,RED);
			else
				PrintColour(con,"OK           ",21,15,GREEN);
	
			// eprom
			if (uniram2 & 4)
				PrintColour(con,"Fault",61,15,RED);
			else
				PrintColour(con,"OK   ",61,15,GREEN);
				
			// eeprom
			if (uniram2 & 8)
				PrintColour(con,"Fault",21,16,RED);
			else
				PrintColour(con,"OK   ",21,16,GREEN);
				
			// ram
			if (uniram2 & 64)
				PrintColour(con,"Fault",61,16,RED);
			else
				PrintColour(con,"OK   ",61,16,GREEN);

			// crankshaft sensor
			if (uniram2 & 128)
				if (uniram3 & 64)
					PrintColour(con,"Short circuit",21,17,RED);
				else
					PrintColour(con,"Open circuit ",21,17,RED);
			else
				PrintColour(con,"OK           ",21,17,GREEN);
		
			// then the immobiliser data	
			ecu = GetECUByte(113);
			if (ecu & 1)
				PrintColour(con,"Ecu in virgin engine mode     ",21,19,RED);
				else
				if (ecu & 2)
					PrintColour(con,"Ecu prohibiting engine start   ",21,19,RED);
					else
						if (ecu & 4)
						PrintColour(con,"Ecu received universal code    ",21,19,RED);
						else
							if (ecu & 16)
							PrintColour(con,"Backdoor activated             ",21,19,RED);
							else
								if (ecu & 32)
								PrintColour(con,"Wrong key code received        ",21,19,RED);
								else
									if (ecu & 64)
									PrintColour(con,"Key code not recognised        ",21,19,RED);
									else
										if (ecu & 128)
										PrintColour(con,"No code received or link failed",21,19,RED);
										else
											PrintColour(con,"OK                             ",21,19,GREEN);
		}	
	}
	
	// finally, tidy up the serial port and the display
	if (rs232 != NULL)
		CloseHandle(rs232);
		
	PrintColour(con,"Bye...\n",0,22,WHITE);
	CloseHandle(con);
	
	// and quit	
	return (0);
}

// provide basic serial system
// including the USB virtual comm port, if any (tested on FTDI232R)

HANDLE serial_setcomms (int baud, char * comport)
{
	// set the comms specs
	if ((rs232 = CreateFile (comport,											// we want the serial port
													GENERIC_READ|GENERIC_WRITE,		// for reading and writing
													0,														// only by us
													NULL,													// no child security attributes
													OPEN_EXISTING,								// because it's the com port
													FILE_ATTRIBUTE_NORMAL,				// we don't want overlapped stuff for single bytes
													NULL))												// no template info allowed
													!= INVALID_HANDLE_VALUE)
	{
		// ok, we opened it!
		// now, see what it looks like
		GetCommState(rs232,&dcb);

		// that has filled the control block with sensible numbers, but we need to change some
		// specifically, it has set the correct size for the size element

		dcb.BaudRate							= baud;	
		dcb.fBinary								= 1;
		dcb.fParity								= 0;										// no parity
		dcb.fOutxCtsFlow					= 0;
		dcb.fOutxDsrFlow					= 0;
		dcb.fDtrControl						= DTR_CONTROL_DISABLE; 
		dcb.fDsrSensitivity				= 0;
		dcb.fTXContinueOnXoff			= 1; 
		dcb.fOutX									= 0;
		dcb.fInX									= 0;
		dcb.fErrorChar						= 0; 
		dcb.fNull									= 0;
		dcb.fRtsControl						= RTS_CONTROL_DISABLE; 
		dcb.fAbortOnError					= 0; 
		dcb.ByteSize							= 8;
		dcb.Parity								= 0;
		dcb.StopBits							= 0;										// one stop bit

		SetCommState(rs232,&dcb);

		// let's see what the time-out functions are...
		GetCommTimeouts(rs232,&cto);
		cto.ReadTotalTimeoutConstant = 100;								// read timeout of 100mS
		SetCommTimeouts(rs232,&cto);
		return rs232;
	}
	else
		return 0;		// on error
}

unsigned char	GetECUByte (unsigned char ch)
{

	// send a character to the engine, wait for the response, and return that response
	// if the loop times out, we return 'noshow'
	
	// 14/7/09
	// to work with 409.1 KKL USB->ISO9141 adaptor, available from
	// ebay seller 'weyri650'
	// requires 'free' cable adaptor or similar to extend the K, L,
	// ground and power to Fiat connector
	//
	// the adaptor uses an interface chip which links the K and L
	// lines internally so everything sent is echoed to the serial
	// port; this byte must be discarded before the actual response
	// byte from the ecu is received.

unsigned char ci;

	// clear the return in case of timeout
	ci = 0;

	PurgeComm(rs232,PURGE_RXABORT	| PURGE_RXCLEAR);

	// send to port
	WriteFile(rs232,&ch,1,&txcount,NULL);
	SleepEx(10,FALSE);

	// read it back
	rxcount = 0;
	// grab the echo byte
	ReadFile(rs232,&ci,1,&rxcount,NULL);
	// grab the return byte
	ReadFile(rs232,&ci,1,&rxcount,NULL);
	if (rxcount == 0)
		ci = 0;
	return (ci);
}

void PrintColour(HANDLE con, char * str, int x, int y, WORD colour)
{
	// print a coloured string at x,y
COORD posn;
	
	posn.X = x;
	posn.Y = y;
	
	SetConsoleTextAttribute(con,colour);
	SetConsoleCursorPosition(con,posn);
	printf("%s",str);
}
	